@@ -1,5 +1,7 @@
 """frame of day generation module"""
+
 import logging
+from typing import Union
 
 import numpy as _np
 import pandas as _pd
@@ -15,7 +17,7 @@ def _get_core_list(core_list_path):
     # need to check if solution numbers are consistent with discontinuities selection
     core_df = _pd.read_csv(
         core_list_path,
-        delim_whitespace=True,
+        sep="\\s+",  # delim_whitespace is deprecated
         skiprows=4,
         comment="-",
         usecols=[0, 1, 2, 3],
@@ -24,10 +26,16 @@ def _get_core_list(core_list_path):
     return core_df
 
 
-def get_frame_of_day(date_or_j2000, itrf_path_or_df, discon_path_or_df, psd_path_or_df=None, list_path_or_df=None):
+def get_frame_of_day(
+    date_or_j2000,
+    itrf_path_or_df: Union[_pd.DataFrame, str],
+    discon_path_or_df: Union[_pd.DataFrame, str],
+    psd_path_or_df: Union[None, _pd.DataFrame, str] = None,
+    list_path_or_df: Union[None, _pd.DataFrame, str, _np.ndarray] = None,
+):
     """Main function to propagate frame into datetime of interest"""
 
-    if isinstance(date_or_j2000, (int, _np.int64)):
+    if isinstance(date_or_j2000, (int, _np.int64)):  # TODO check: np.int64 is meant to be a class not a type
         date_J2000 = date_or_j2000
     else:
         date_J2000 = _gn_datetime.datetime2j2000(_np.datetime64(date_or_j2000))
@@ -38,37 +46,37 @@ def get_frame_of_day(date_or_j2000, itrf_path_or_df, discon_path_or_df, psd_path
     elif isinstance(discon_path_or_df, str):
         discon_df = _gn_io.discon._read_discontinuities(discon_path_or_df)
     else:
-        logging.error(msg="check discon_path_or_df")
+        raise ValueError(f"discon_path_or_df must be a pandas DataFrame or str, got: {type(discon_path_or_df)}")
 
     # itrf sinex file
     if isinstance(itrf_path_or_df, _pd.DataFrame):
         output = itrf_path_or_df
     elif isinstance(itrf_path_or_df, str):
         output = _gn_io.sinex._get_snx_vector_gzchunks(filename=itrf_path_or_df, block_name="SOLUTION/ESTIMATE")
+        if output is None:
+            raise Exception(f"Output from _get_snx_vector_gzchunks() was None! Filepath: {itrf_path_or_df}")
     else:
-        logging.error(msg="check itrf_path_or_df")
+        raise ValueError(f"itrf_path_or_df must be a pandas DataFrame or str, got: {type(itrf_path_or_df)}")
 
-    discon_valid = discon_df[
-        (discon_df.MODEL == "P")
-        & (discon_df.START <= date_J2000)
-        & (discon_df.END > date_J2000)
-    ]
+    discon_valid = discon_df[(discon_df.MODEL == "P") & (discon_df.START <= date_J2000) & (discon_df.END > date_J2000)]
 
-    comboindex = _pd.Index(output.CODE_PT + '_' + output.SOLN.values.astype(str))
-    itrf_code_pt = _pd.Index(output.CODE_PT.values, name=None).unique()
-    soln_series = _pd.Series(data=0,index=itrf_code_pt,dtype=int)
+    itrf_code_pt_index = output.index.get_level_values("CODE_PT")
+    comboindex = itrf_code_pt_index + "_" + output.index.get_level_values("SOLN").astype("str")
+    itrf_code_pt_uniques = itrf_code_pt_index.unique()
+    soln_series = _pd.Series(data=0, index=itrf_code_pt_uniques, dtype=int)
 
     # clean discontinuities file from stations missing in the frame
     comboindex_dv = _pd.Index(discon_valid.CODE.values + "_" + discon_valid.PT.values)
-    dv_mask = comboindex_dv.isin(itrf_code_pt)
+    dv_mask = comboindex_dv.isin(itrf_code_pt_uniques)
 
     # overwrite indices on discont
     soln_series.loc[comboindex_dv[dv_mask]] = discon_valid[dv_mask].SOLN
-    soln_series = soln_series[soln_series.values.astype(bool)] # remove all the stations that do not have a valid solution
+    # remove all the stations that do not have a valid solution
+    soln_series = soln_series[soln_series.values.astype(bool)]
     if not soln_series.values.nonzero()[0].size:
         logging.error(msg="list stations are not in frame sinex")
-    #TODO cite specific page/line of the standard
-    soln_comboindex = soln_series.index.values + '_' + soln_series.values.astype(str)
+    # TODO cite specific page/line of the standard
+    soln_comboindex = soln_series.index.values + "_" + soln_series.values.astype(str)
     out = output[comboindex.isin(soln_comboindex)]
 
     if list_path_or_df is not None:
@@ -78,21 +86,27 @@ def get_frame_of_day(date_or_j2000, itrf_path_or_df, discon_path_or_df, psd_path
         elif isinstance(list_path_or_df, str):
             core_df = _get_core_list(list_path_or_df)
             core_list = core_df.CODE.values
-        elif isinstance(list_path_or_df, _np.ndarray) or isinstance(list_path_or_df, list):
+        elif isinstance(list_path_or_df, (_np.ndarray, list)):
             core_list = list_path_or_df
         else:
-            logging.error(msg="check list_path_or_df")
-        out_mask = out.CODE_PT.str[:4].isin(core_list)
+            raise ValueError(
+                f"list_path_or_df must be a pandas DataFrame, str or numpy ndarray, got: {type(list_path_or_df)}"
+            )
+        out_mask = out.index.get_level_values("CODE_PT").str[:4].isin(core_list)
 
-        if out_mask.values.any():
+        if out_mask.any():
             out = out[out_mask]
         else:
             logging.error(msg="list stations not present in frame file")
             return None
 
-    out_xyzvel = _pd.Series(data = out.VAL.EST.values,index = _pd.MultiIndex.from_arrays([out.CODE_PT.values, out.TYPE])).unstack(1)
+    out_xyzvel = _pd.Series(
+        data=out.VAL.EST.values,
+        index=_pd.MultiIndex.from_arrays([out.index.get_level_values("CODE_PT"), out.index.get_level_values("TYPE")]),
+    ).unstack(1)
 
-    itrf_reference = out["REF_EPOCH"].values[0] # all values in the frame are relative to frame origin reference time (2010-01-02T00:00:00)
+    # all values in the frame are relative to frame origin reference time (2010-01-02T00:00:00)
+    itrf_reference = out.index.get_level_values("REF_EPOCH")[0]
     time_seconds = date_J2000 - itrf_reference
 
     position = out_xyzvel.iloc[:, :3]
@@ -107,10 +121,10 @@ def get_frame_of_day(date_or_j2000, itrf_path_or_df, discon_path_or_df, psd_path
             logging.info(msg="reading postseismic from the dataframe provided")
             psd_df = psd_path_or_df
         elif isinstance(psd_path_or_df, str):
-            logging.info(msg="reading postseismic from {psd_path_or_df}")
+            logging.info(msg=f"reading postseismic from {psd_path_or_df}")
             psd_df = _gn_io.psd._get_psd_df(psd_path_or_df)
         else:
-            logging.error(msg="check psd_path_or_df")
+            raise ValueError(f"psd_path_or_df must be a pandas DataFrame or str, got: {type(psd_path_or_df)}")
         logging.info(msg="applying the computed postseismic deformations")
         out = psd2frame(frame_of_day=out, psd_df=psd_df)
     return out
@@ -128,7 +142,7 @@ def psd2frame(frame_of_day, psd_df):
     # select only those sites that have psd event
     psd_enu = psd_enu[psd_enu.any(axis=1)]
 
-    llh = _gn_transform.xyz2llh_heik(xyz_array=frame_of_day[["STAX", "STAY", "STAZ"]].loc[psd_enu.index].values)
+    llh = _gn_transform.xyz2llh(xyz_array=frame_of_day[["STAX", "STAY", "STAZ"]].loc[psd_enu.index].values)
     phi, lam = llh[:, 0], llh[:, 1]
     rot = _gn_transform.llh2rot(phi=phi, lamb=lam, enu_to_ecef=True)
 
@@ -158,15 +172,17 @@ def _get_psd_enu(psd_df, date_J2000):
 
     out = log_part_grouped.add(exp_part_grouped, fill_value=0) / 1000
     # if log or exp part is missing for the component, .add should take care of the nans being added
-    # .rename(columns={'E':'EAST','N':'NORTH','H':'UP'})
-    return out[["E", "N", "H"]]
+    out.rename(columns={"E": "EAST", "N": "NORTH", "H": "UP", "U": "UP"}, inplace=True, errors="ignore")
+    return out[["EAST", "NORTH", "UP"]]
 
 
 # gather_reader
 def read_frame_snx_all(*file_paths, core_sites=None):
     buf = []
     for path in file_paths:
-        buf.append(_gn_io.sinex._get_snx_vector_gzchunks(filename=path, block_name="SOLUTION/ESTIMATE"),snx_format='raw')
+        buf.append(
+            _gn_io.sinex._get_snx_vector_gzchunks(filename=path, block_name="SOLUTION/ESTIMATE"), snx_format="raw"
+        )
     all_frame = _pd.concat(buf)
     if core_sites is not None:
         return all_frame[all_frame.CODE.isin(core_sites)]
