@@ -1,11 +1,19 @@
+import datetime
+import math
 import os
-from collections.abc import Callable, Iterable
-from io import BytesIO as _BytesIO
-from numbers import Real
-from typing import List, TextIO, Union
+import pathlib
 
+# The collections.abc (rather than typing) versions don't support subscripting until 3.9
+# from collections.abc import Callable, Iterable
+from typing import Callable, Iterable
+from io import BytesIO as _BytesIO
+from typing import List, TextIO, Union
+from urllib import request as _rqs
+
+import numpy as _np
 import pandas as _pd
 
+from gnssanalysis import gn_datetime as _gn_datetime
 from gnssanalysis import gn_io as _gn_io
 
 
@@ -44,7 +52,7 @@ def merge_hyphen_headers(raw_split_header: Iterable[str]) -> List[str]:
     This function re-merges those header components.
 
     :param Iterable[str] raw_split_header: ERP header line that has been split/tokenized
-    :return List[str]: List of ERP headers with hyphen-seperated headers merged
+    :return List[str]: List of ERP headers with hyphen-separated headers merged
     """
     # Copy to avoid mutating input list
     headers = list(raw_split_header)
@@ -78,11 +86,11 @@ def merge_hyphen_headers(raw_split_header: Iterable[str]) -> List[str]:
 def normalise_stddevcorr_headers(header: str) -> str:
     """Create a canonical representation of (partially normalised) std.dev or corr ERP headers
 
-    This canonicisation process involves stripping a leading "S" from a header, replacing it with
-    a trailing "SIG", stripping a leading "C" and replacing it with a trainling "CORR" and changing
+    This canonicalisation process involves stripping a leading "S" from a header, replacing it with
+    a trailing "SIG", stripping a leading "C" and replacing it with a trailing "CORR" and changing
     a "COR" suffix to be "CORR".
 
-    :param str header: paritially normalised ERP header
+    :param str header: partially normalised ERP header
     :return str: ERP header with canonical std.dev and corr. representation
     """
     if header[0] == "S":
@@ -194,7 +202,11 @@ def get_erp_unit_string(normalised_header: str, original_header: str) -> str:
         return "E-6as/d"
     elif normalised_header in ["UT1-UTC", "UT1R-UTC", "UT1-TAI", "UT1R-TAI", "UTsig", "UTRsig"]:
         return "E-7s"
-    elif normalised_header in ["LOD", "LODR", "LODsig", ]:
+    elif normalised_header in [
+        "LOD",
+        "LODR",
+        "LODsig",
+    ]:
         return "E-7s/d"
     elif normalised_header in ["Deps", "Depssig", "Dpsi", "Dpsisig"]:
         return "E-6"
@@ -216,7 +228,7 @@ def read_erp(
     :param bool normalise_header_names: If True, change header names to canonical versions, defaults to True
     :param bool convert_units: Convert natural ERP file units to "SI units", forces `normalise_header_names`, defaults to True
     :raises RuntimeError: Raised if the start of the column headers can't be found and the file can't be parsed
-    :return _pd.DataFrame: A dataframe containing all the columnar data from the ERP file as well as an `attrs` dict
+    :return _pd.DataFrame: A DataFrame containing all the columnar data from the ERP file as well as an `attrs` dict
         holding the header comments in a "header_comments" property and whether the data has been scaled in a "scaled"
         property
     """
@@ -248,7 +260,7 @@ def read_erp(
     data_of_interest = content[start_of_data:]  # data block
     erp_df = _pd.read_csv(
         _BytesIO(data_of_interest),
-        delim_whitespace=True,
+        sep="\\s+",  # delim_whitespace is deprecated
         names=headers,
         index_col=False,
     )
@@ -271,7 +283,7 @@ def read_erp(
     return erp_df
 
 
-def get_erp_column_formatter(column_header: str, mjd_precision: int = 2) -> Callable[[Real], str]:
+def get_erp_column_formatter(column_header: str, mjd_precision: int = 2) -> Callable[[float], str]:
     """Get an appropriate formatter for the data a column of ERP data
 
     Different ERP columns are formatted slightly differently and so this function provides access to
@@ -279,8 +291,8 @@ def get_erp_column_formatter(column_header: str, mjd_precision: int = 2) -> Call
     of decimal points for the MJD to account for PEA's non-standard high-rate ERP data.
 
     :param str column_header: Normalised ERP column header
-    :param int mjd_precision: Numer of decimal places to use for MJD column, defaults to 2
-    :return Callable[[Real], str]: Formatting function that can be applied to values in a column
+    :param int mjd_precision: Number of decimal places to use for MJD column, defaults to 2
+    :return Callable[[float], str]: Formatting function that can be applied to values in a column
     """
     if column_header == "MJD":
         return lambda val: format(val, f".{mjd_precision}f")
@@ -290,7 +302,7 @@ def get_erp_column_formatter(column_header: str, mjd_precision: int = 2) -> Call
         return lambda val: format(int(val), "d")
 
 
-def format_erp_column(column_series: _pd.Series, mjd_precision: int = 2) -> _pd.Series[str]:
+def format_erp_column(column_series: _pd.Series, mjd_precision: int = 2) -> _pd.Series:
     """Formats an ERP DataFrame column into appropriate strings
 
     Includes the ability to specify a non-standard number of decimal points for MJD to account
@@ -298,7 +310,7 @@ def format_erp_column(column_series: _pd.Series, mjd_precision: int = 2) -> _pd.
 
     :param _pd.Series column_series: DataFrame column to format into strings
     :param int mjd_precision: Number of decimal places to use for MJD column, defaults to 2
-    :return _pd.Series[str]: Series (DataFrame column) full of formatted values
+    :return _pd.Series: Series (DataFrame column) full of formatted values
     """
     formatter = get_erp_column_formatter(str(column_series.name), mjd_precision)
     return column_series.apply(formatter)
@@ -345,6 +357,245 @@ def write_erp_to_stream(erp_df: _pd.DataFrame, stream: TextIO, mjd_precision: in
     stream.write(" ".join(s.rjust(w) for (s, w) in zip(unit_strings, column_widths)))
     stream.write("\n")
     # Values
-    for (_, row) in value_strings_df.iterrows():
+    for _, row in value_strings_df.iterrows():
         stream.write(" ".join(s.rjust(w) for (s, w) in zip(row, column_widths)))
         stream.write("\n")
+
+
+def read_iau2000(iau2000_path: Union[str, bytes, os.PathLike], use_erp_style_headers: bool = True) -> _pd.DataFrame:
+    """Read an IAU2000 file from disk into a pandas DataFrame
+
+    All columns of data are preserved, included the IERS/Predicted markers. Where data is absent in the IAU2000 file
+    a NaN is placed into the pandas DataFrame. The returned DataFrame can either be provided with ERP style headers,
+    eg. Xpole for polar motion, or titles that align closer to the IERS description of the data, eg. PM-x.
+
+    :param Union[str, bytes, os.PathLike] iau2000_path: Path to IAU2000 file to read
+    :param bool use_erp_style_headers: Use headers that align with ERP column names, defaults to True
+    :return _pd.DataFrame: A pandas DataFrame containing the data in the IAU2000 file
+    """
+    iau2000_bytes = _gn_io.common.path2bytes(str(iau2000_path))
+    # See https://maia.usno.navy.mil/ser7/readme.finals2000A
+    # Note for colspecs that the above reference is [closed, closed] intervals
+    # for 1-indexed columns, colspecs takes [closed, open] intervals for 0-indexed
+    # columns
+    iau2000_df = _pd.read_fwf(
+        _BytesIO(iau2000_bytes),
+        colspecs=[
+            (0, 2),  # Year
+            (2, 4),  # Month
+            (4, 6),  # Day
+            (7, 15),  # MJD
+            (16, 17),  # IERS/Prediction flag polar motion (A)
+            (18, 27),  # Xpole (A)
+            (27, 36),  # Xsig (A)
+            (37, 46),  # Ypole (A)
+            (46, 55),  # Ysig (A)
+            (57, 58),  # IERS/Prediction flag UT1-UTC (A)
+            (58, 68),  # UT1-UTC (A)
+            (68, 78),  # UTsig (A)
+            (79, 86),  # LOD (A)
+            (86, 93),  # LODsig (A)
+            (95, 96),  # IERS/Prediction flag nutation (A)
+            (97, 106),  # Xrt (A)
+            (106, 115),  # Xrtsig (A)
+            (116, 125),  # Yrt (A)
+            (125, 134),  # Yrtsig (A)
+            (134, 144),  # Xpole (B)
+            (144, 154),  # Ypole (B)
+            (154, 165),  # UT1-UTC (B)
+            (165, 174),  # Xrt (B)
+            (175, 184),  # Yrt (B)
+        ],
+        header=None,
+    )
+    if use_erp_style_headers:
+        iau2000_df.columns = [
+            "Year",
+            "Month",
+            "Day",
+            "MJD",
+            "MotionIPflag",
+            "Xpole",
+            "Xsig",
+            "Ypole",
+            "Ysig",
+            "UT1IPflag",
+            "UT1-UTC",
+            "UTsig",
+            "LOD",
+            "LODsig",
+            "RateIPflag",
+            "Xrt",
+            "Xrtsig",
+            "Yrt",
+            "Yrtsig",
+            "Xpole-B",
+            "Ypole-B",
+            "UT1-UTC-B",
+            "Xrt-B",
+            "Yrt-B",
+        ]
+    else:
+        iau2000_df.columns = [
+            "Year",
+            "Month",
+            "Day",
+            "MJD",
+            "PM_IP",
+            "PM-x",
+            "error_PM-x",
+            "PM-y",
+            "error_PM-y",
+            "UT1_IP",
+            "UT1-UTC",
+            "error_UT1-UTC",
+            "LOD",
+            "error_LOD",
+            "dX_IP",
+            "dX",
+            "error_dX",
+            "dY",
+            "error_dY",
+            "PM-x-B",
+            "PM-y-B",
+            "UT1-UTC-B",
+            "dX-B",
+            "dY-B",
+        ]
+    return iau2000_df
+
+
+def iau2000_df_to_erp_df(iau2000_df: _pd.DataFrame, erp_units: bool = False) -> _pd.DataFrame:
+    """Transform an IAU2000 DataFrame into an equivalent ERP-style DataFrame
+
+    This process can only be applied to an IAU2000 DataFrame that has ERP-style column headers.
+    The transformation mostly consists of dropping many of the IAU2000 columns, adding empty Nt, Nf, Nr
+    columns, and transforming the data to appropriate ERP units. This can either be "native ERP units"
+    or "SI units", as per :func: `gn_io.erp.read_erp` and both are compatible with writing to disk.
+
+    :param _pd.DataFrame iau2000_df: IAU2000 DataFrame
+    :param bool erp_units: If False, output is in "SI units", if true output is in "native ERP units", defaults to False
+    :return _pd.DataFrame: ERP DataFrame containing data equivalent to IAU2000 DataFrame
+    """
+    erp_df = _pd.DataFrame(
+        columns=[
+            "MJD",
+            "Xpole",
+            "Ypole",
+            "UT1-UTC",
+            "LOD",
+            "Xsig",
+            "Ysig",
+            "UTsig",
+            "LODsig",
+            "Xrt",
+            "Yrt",
+            "Xrtsig",
+            "Yrtsig",
+        ]
+    )
+    for col in erp_df:
+        erp_df[col] = iau2000_df[col] * get_iau2000_to_erp_scaling(col, erp_units)
+
+    erp_df.insert(loc=9, column="Nt", value=_np.zeros(len(erp_df)))
+    erp_df.insert(loc=9, column="Nf", value=_np.zeros(len(erp_df)))
+    erp_df.insert(loc=9, column="Nr", value=_np.zeros(len(erp_df)))
+    erp_df.attrs["scaled"] = not erp_units
+    erp_df.attrs["header_comments"] = (
+        # Using adjacent string concatenation to build multiline string
+        # Currently we're just stealing the IGU header, review this later
+        "Source: Xpole,Ypole,Xrt,Yrt,LOD: weighted average of centres;\n"
+        "        UT1-UTC: integrated from the 5th day prior to Bull. A\n"
+        "                 last non-predicted value.\n"
+        "\n"
+        "Orbits: to be used with the IGS Ultra Rapid Orbits (IGU)\n"
+        "\n"
+    )
+    return erp_df
+
+
+def get_iau2000_to_erp_scaling(column_header: str, erp_units: bool) -> float:
+    """Given an ERP column header, return the scaling factor from IAU2000 data to ERP data
+
+    The scaling is such that ERP = scaling * IAU2000. In this context "ERP" can also be either
+    "SI units" or "ERP native units" in line with :func: `gn_io.erp.read_erp`.
+
+    :param str column_header: ERP column name to get scaling for
+    :param bool erp_units: If False, scaling is to "SI units", if true scaling is to "native ERP units"
+    :return float: Scaling factor from IAU2000 units to ERP units
+    """
+    iau2000_scaling = get_iau2000_scaling(column_header)
+    if erp_units:
+        erp_scaling = get_erp_scaling(column_header, column_header)
+    else:
+        erp_scaling = 1
+    return iau2000_scaling / erp_scaling
+
+
+def get_iau2000_scaling(column_header: str) -> Union[int, float]:
+    """Given an IAU2000 column header, return the scaling factor from IAU2000 data to "SI units"
+
+    Namely the scaling shifts LOD properties from milliseconds to seconds and polar motion rate properties
+    from milliarcseconds to arcseconds.
+
+    :param str column_header: IAU2000 column header in either ERP-style or IERS-style
+    :return Union[int, float]: Scaling factor to go from IAU2000 units to "SI units"
+    """
+    if column_header in ["PM-x", "Xpole", "PM-y", "Ypole", "PM-x-B", "Xpole-B", "PM-y-B", "Ypole-B"]:
+        return 1
+    elif column_header in ["error_PM-x", "Xsig", "error_PM-y", "Ysig"]:
+        return 1
+    elif column_header in ["UT1-UTC", "UT1-UTC-B", "error_UT1-UTC", "UTsig"]:
+        return 1
+    elif column_header in ["LOD", "error_LOD", "LODsig"]:
+        return 1e-3
+    elif column_header in ["dX", "Xrt", "dY", "Yrt", "dX-B", "Xrt-B", "dY-B", "Yrt-B"]:
+        return 1e-3
+    elif column_header in ["error_dX", "Xrtsig", "error_dY", "Yrtsig"]:
+        return 1e-3
+    else:
+        return 1
+
+
+def erp_outfile(datetime_epoch: datetime.datetime, output_dir: pathlib.Path):
+    """
+    Input datetime string of format "YY-MM-DD hh:mm:ss"
+    """
+    mjd = _gn_datetime.pydatetime_to_mjd(datetime_epoch)
+
+    # Download the IAU2000 daily finals file
+    if pathlib.Path("finals.daily.iau2000.txt").is_file():
+        pathlib.Path("finals.daily.iau2000.txt").unlink()
+    iers_url = "https://datacenter.iers.org/products/eop/rapid/daily/finals2000A.daily"
+    iau2000_daily_file = pathlib.Path.cwd() / "finals.daily.iau2000.txt"
+    _rqs.urlretrieve(iers_url, filename=iau2000_daily_file)
+    # Read the data in the IAU2000 file
+    iau2000_df = read_iau2000(iau2000_daily_file)
+    # Delete the file on disk
+    iau2000_daily_file.unlink()
+
+    # Filter to the region of time we care about and transform to ERP DF
+    times_of_interest_mask = (iau2000_df["MJD"] > mjd - 10) & (iau2000_df["MJD"] < mjd + 3.1)
+    erp_df = iau2000_df_to_erp_df(iau2000_df[times_of_interest_mask].dropna())
+
+    erp_df = erp_df.set_index("MJD")
+    first_half_day = math.floor(erp_df.first_valid_index()) + 0.5
+    max_day = math.ceil(erp_df.last_valid_index())
+    desired_output_index = _pd.Index(data=_np.arange(first_half_day, max_day, step=1))
+    combined_index = erp_df.index.union(desired_output_index)
+
+    resampled_df = (
+        erp_df.reindex(index=combined_index)
+        .interpolate(method="quadratic", fill_value="extrapolate")
+        .reindex(index=desired_output_index)
+        .reset_index(names="MJD")
+    )
+
+    # Restrict the amount data we output
+    resampled_df = resampled_df[resampled_df["MJD"] > mjd - 3]
+
+    gps_date = _gn_datetime.gpsweekD(datetime_epoch.strftime("%Y"), datetime_epoch.strftime("%j"), wkday_suff=True)
+    file_suffix = f'_{int(int(str(mjd).split(".")[1].ljust(2,"0"))*0.24):02}'
+    file_name = f"igu{gps_date}{file_suffix}.erp"
+
+    write_erp(resampled_df, output_dir / file_name)
